#!/usr/bin/perl

# TODO: template profiles to support different types of tagging (address device deficiencies)
# TODO: useful statistics at end of encoding

use strict;
use warnings;

use open ':utf8';

use Encode;
use File::Find qw(find);
use File::Path qw(mkpath);
use File::Spec::Functions qw(catpath rel2abs splitpath);
use Getopt::ArgvFile home=>1;
use Getopt::Long qw(GetOptions);
use Parallel::ForkManager;
use Sys::CPU;

# name of this program and associated version
my $progname = 'flac2mp3';
my $version = '0.9';

# command line options (defaulted)
my $encopts = '--preset standard --noreplaygain';
my $flac = 'flac';
my $lame = 'lame';
my $metaflac = 'metaflac';
my $processes = Sys::CPU::cpu_count();

# command line options (undefined)
my $quiet; # false
my $overwrite; # false
my $verbose; # false

# command line arguments
my $flac_dir;
my $mp3_dir;

# mappings of common vorbis tags to id3v2.3 frames (with priorities)
my %tag_mappings = (
  'ALBUM'               => 'TALB:0',  # album/movie/show title
  'ALBUMARTIST'         => 'TPE2:1',  # band/orchestra/accompaniment
  'ARRANGER'            => 'TPE4:1',  # interpreted, remixed, or otherwise modified by
  'ARTIST'              => 'TPE1:1',  # lead performer(s)/soloist(s)
  'AUTHOR'              => 'TEXT:2',  # lyricist/text writer
  'BPM'                 => 'TBPM:0',  # BPM (beats per minute)
  'COMMENT'             => 'COMM:0',  # comments
  'COMPOSER'            => 'TCOM:0',  # composer
  'CONDUCTOR'           => 'TPE3:0',  # conductor/performer refinement
  'COPYRIGHT'           => 'TCOP:0',  # copyright message
  'DATE'                => 'TYER:0',  # year
  'DISCNUMBER'          => 'TPOS:0',  # part of a set
  'ENCODEDBY'           => 'TENC:0',  # encoded by
  'ENSEMBLE'            => 'TPE2:2',  # band/orchestra/accompaniment
  'GENRE'               => 'TCON:0',  # content type
  'GROUPING'            => 'TIT1:0',  # content group description
  'ISRC'                => 'TSRC:0',  # ISRC (international standard recording code)
  'LABEL'               => 'TPUB:2',  # publisher
  'LYRICIST'            => 'TEXT:1',  # lyricist/text writer
  'ORGANIZATION'        => 'TPUB:3',  # publisher
  'ORIGINALALBUM'       => 'TOAL:0',  # original album/movie/show title
  'ORIGINALARTIST'      => 'TOPE:0',  # original artist(s)/performer(s)
  'ORIGINALDATE'        => 'TORY:0',  # original release year
  'LYRICS'              => 'USLT:0',  # unsynchronized lyric/text transcription
  'MUSICBRAINZ_TRACKID' => 'UFID:0',  # unique file identifier
  'PART'                => 'TIT3:1',  # subtitle/description refinement
  'PERFORMER'           => 'TPE1:2',  # lead performer(s)/soloists(s) 
  'PUBLISHER'           => 'TPUB:1',  # publisher
  'REMIXER'             => 'TPE4:2',  # interpreted, remixed, or otherwise modified by
  'SUBTITLE'            => 'TIT3:2',  # subtitle/description refinement
  'TRACKID'             => 'UFID:1',  # unique file identifier
  'TITLE'               => 'TIT2:0',  # title/songname/content description
  'TRACKNUMBER'         => 'TRCK:0',  # track number/position in set
  'VERSION'             => 'TIT3:3',  # subtitle/description refinement
  'WEBSITE'             => 'WOAR:0',  # official artist/performer webpage
);

# return true if value defined and not empty
sub is_value {
  return (defined($_[0]) && length($_[0]) > 0);
}

my $header; # false

sub header {

  # do nothing if already output or user asked to be quiet
  return if (defined($header) || defined($quiet));

  print STDERR "\n";
  print STDERR "$progname $version: Transcodes FLAC files to MP3 format en masse.\n";
  print STDERR "Copyright 2006-2009 Paul C. Bryan.\n";
  print STDERR "\n";

  $header = 1;
}

# output informative message if not quiet
sub msg {
  header; print STDERR "$_[0]\n" if (!defined($quiet));
}

# display error message if not quiet, then exit with error code
sub error {
  undef $quiet; msg("$_[0]\n"); exit 1;
}

# output diagnostic if verbosity enabled
sub diag {
  msg($_[0]) if (defined($verbose));
}

# display system call as diagnostic, then execute system call
sub sys {
  diag($_[0]); system($_[0]) == 0 or error("Error calling $_[0].");
}

# trims spaces from start and end of string
sub trim {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  $v =~ s/^[[:space:]]+//; $v =~ s/[[:space:]]+$//;
  return $v;
}

# escape special characters to make appropriate for quoted program arguments
sub escq {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  $v =~ s/\"/\\\"/g;
  $v =~ s/\$/\\\$/g;
  return '"' . $v . '"';
}

sub trail_slash {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  $v =~ s/([^\/])$/$1\//;
 return $_[0];
}

# returns "no" if boolean false (undefined)
sub noboolean {
  return (!defined($_[0]) ? "no" : "");
}

# display usage and exit
sub usage {

  undef $quiet;
  header();

	print STDERR <<EOF;
Usage: $progname [options] flac_dir [mp3_dir]

  --help             Display this help message.
  --encopts=<opts>   Options to pass to LAME for encoding [$encopts].
  --flac=<prog>      Program to decode extracted audio from FLAC format [$flac].
  --lame=<prog>      Program to encode to MP3 format [$lame].
  --metaflac=<prog>  Program to get metadata from FLAC files [$metaflac].
  --[no]overwrite    Overwrite previous extract of tracks [${\(noboolean($overwrite))}overwrite].
  --processes=<num>  Number of concurrent encoder processes [$processes].
  --quiet            Quiet operation, no screen output.
  --verbose          Verbose output of progress information.

CONFIGURATION FILE:
  You can store persistent command-line options in a file called .$progname in
  your home directory. They will be read as defaults prior to processing any
  passed command-line options.

EOF

	exit 1;
}

# parse command line options (may override defaults or initialization)
sub get_options {

  GetOptions(
    'encopts=s'   => \$encopts,
    'flac=s'      => \$flac,
    'help'        => \&usage,
    'overwrite'   => sub { $overwrite = 1; },
    'lame=s'      => \$lame,
    'metaflac=s'  => \$metaflac,
    'nooverwrite' => sub { undef $overwrite; },
    'processes=i' => \$processes,
    'quiet'       => sub { $quiet = 1; undef $verbose; },
    'verbose'     => sub { $verbose = 1; undef $quiet; },
  ) or usage;

  # first argument mandatory, second optional
  usage() if (@ARGV < 1 || @ARGV > 2);

  # first parameter: directory to find flac files (make absolute)
  $flac_dir = rel2abs($ARGV[0]);

  # second (optional, defaults to flac directory)
  $mp3_dir = (@ARGV >= 2 ? rel2abs($ARGV[1]) : $flac_dir);

  # ensure director(ies) specified are indeed directories
  error("$flac_dir is not a directory.") if (!-d $flac_dir);
  error("$mp3_dir is not a writeable directory.") if !(-d $mp3_dir && -w $mp3_dir);

  # directories should have trailing slashes
  $flac_dir = trail_slash($flac_dir);
  $mp3_dir = trail_slash($mp3_dir);
}

# get metadata from flac file
sub get_metadata {

  my $flac = $_[0];

  my %metadata = ();

  open(METAFLAC, "$metaflac --export-tags-to=- " . escq($flac) . ' |');

  while (<METAFLAC>) {
  
    chomp;

# FIXME: multiple tags of same name overwritten...
    # assign metadata value to appropriate hash entry
    if (m/^(.*)=(.*)/) {
     $metadata{$1} = $2 if (is_value($1) && is_value($2));
    }
  }

  close(METAFLAC);

  return %metadata;
}

# extract year from date
sub extract_year {
  return undef if (!defined($_[0]));
  my $v = $_[0];
  if ($v =~ m/(\d{4})/) { return $v; }
  return undef;
}

# munge the metadata so that mp3 file metadata encoding is best
sub munge_metadata {

  my $metadata = $_[0];

  # truncate date into just its year
  my $date = $metadata->{'DATE'};
  $metadata->{'DATE'} = extract_year($date) if (is_value($date));

  my $title = $metadata->{'TITLE'};
  my $featuring = $metadata->{'FEATURING'};
  $title .= " (feat. $featuring)" if (is_value($title) && is_value($featuring));
  my $version = $metadata->{'VERSION'};
  $title .= " ($version)" if (is_value($title) && is_value($version));
  $metadata->{'TITLE'} = $title if (is_value($title));
}

sub process_file {

  # get the complete pathname to the file
  my $flac_file = $_[0];

  # get flac file path relative to specified flac directory
  my $relative_flac = substr($flac_file, length($flac_dir));

  # create mp3 file path from flac relative file path
  my $mp3_path = $relative_flac;
  $mp3_path =~ s/(.flac$)/.mp3/;
  $mp3_path = $mp3_dir . $mp3_path;
  
  # do not overwrite unless explicitly requested
  if (-e $mp3_path && !defined $overwrite) {
    diag("Ignoring $mp3_path (not overwriting existing file).");
    return;
  }

  # a most useful message
  msg("Writing $mp3_path.");

  # get metadata from the file
  my %metadata = get_metadata($flac_file);

  munge_metadata(\%metadata);

  # get base directory for mp3 file
  my ($vol, $dir, $file) = splitpath($mp3_path);
  my $mp3_base = catpath($vol, $dir, '');
  
  # make base directory if it does not exist
  eval { mkpath($mp3_base) };
  if ($@) { error("Couldn't create $mp3_base directory for files ($@)."); }

  # transcode the audio data
  sys("$flac" .
   ((!defined $verbose) ? ' --silent' : '') .
   ' --decode --stdout ' . escq($flac_file) .
   " | $lame " .
   ((!defined $verbose) ? ' --quiet' : '') .
   ((defined $verbose) ? ' --verbose' : '') . " $encopts" .
   ' - ' . escq($mp3_path));

  my %frames;
  
  while (my ($key, $value) = each %metadata) {
  
    next if (!is_value($key) || !is_value($value));
    my $mapping = $tag_mappings{$key};
    next if (!is_value($mapping));
    my ($frame, $priority) = split(/:/, $mapping);
    
    if (!defined($frames{$frame}) || $priority lt substr($frames{$frame}, 0, 1)) {
      $frames{$frame} = "$priority$value";
    }
  }

  my $id3v2;

  while (my ($frame, $value) = each %frames) {
    next if (!is_value($frame) || !is_value($value));
    $value = substr($value, 1);
    next if (!is_value($value));
    $id3v2 .= " --$frame " . escq($value);
  }

  # convert to iso-8859 octets
  $id3v2 = encode('iso-8859-1', $id3v2);

  # set metadata in mp3 track
  sys('id3v2' . $id3v2 . ' ' . escq($mp3_path));
}

# place to store found files
my @files = ();

sub push_file {

  # get the complete pathname to the file
  my $file = $_;

  # if file is a directory, ignore it
  return if (-d $file);

  # if file does not have flac extension, ignore it
  return if (!($file =~ m/\.flac$/));

  # file must be in flac directory because we're eventually going to compute relative name
  error("$file not in $flac_dir") if (index($file, $flac_dir) != 0);

  push(@files, $file);
}

# iterate through files in flac directory, calling process_file for each
sub process_files {

  # get all files up-front
  find({wanted =>\&push_file, no_chdir=> 1, follow=>1, follow_skip=>2, dangling_symlinks=>0}, $flac_dir);

  # some deterministic way of iterating through the files found
  @files = sort { $a cmp $b } @files;

  my $pm = new Parallel::ForkManager($processes);

  for my $file (@files) {
    $pm->start() and next;
    process_file($file);
    $pm->finish();
  }

  $pm->wait_all_children();
}

# output program header
header();

# parse command line options
get_options();

# iterate through all files in the flac directory, processing each
process_files();

msg "\nDone.";
